/**
 * Copyright (c) 2014, nodebundle.js author(s)
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 *  purpose with or without fee is hereby granted, provided that the above
 *  copyright notice and this permission notice appear in all copies.
 */

var fs   = require('fs');
var path = require('path');

function hereDoc(f) {
	return f.toString().slice(16,-3);
}

function longestCommonPrefix(strings) {
	if (! strings.length) return '';
	
	var minlen = strings[0].length;
	var i;
	
	for (i = 0; i !== strings.length; ++i) {
		minlen = Math.min(minlen, strings[0].length);
	}
	
	for (i = 0; i !== minlen; ++i) {
		var char = strings[0][i];
		for (var j = 1; j !== strings.length; ++j) {
			if (strings[j][i] !== char) {
				// i'th position doesn't match.
				return strings[j].substr(0, i);
			}
		}
	}
	
	// All positions match up to minlen.
	return strings[0].substr(0, minlen+1);
}

//
// Bundle class
//

function Bundle() {
	this.files = [];
	this.execute = [];
}

Bundle.prototype.addDir = function(dir) {
	var files = fs.readdirSync(dir);
	for (var i = 0; i !== files.length; ++i) {
		
		var fullname = path.join(dir, files[i]);
		
		if (fs.statSync(fullname).isDirectory()) {
			this.addDir(fullname);
			
		} else if ( !! fullname.match(/\.js$/i)) {
			this.addFile(fullname);
			
		}
		
	}
};

Bundle.prototype.addFile = function(filepath) {
	// Normalise directory separators to what our require() implementation uses
	var identifier = this.normalisePath(filepath);
	
	var filebody = fs.readFileSync(filepath);
	
	// nullify leading #!/usr/bin/shebang if present
	
	if (filebody.slice(0, 2) == "#!") { // apparently === doesn't work
		filebody = "//" + filebody;
	}
	
	this.files[identifier] = filebody;
};

Bundle.prototype.addExecute = function(filepath) {
	this.execute.push( this.normalisePath(filepath) );
};

Bundle.prototype.pruneByRegexString = function(str) {
	var r = new RegExp(str);
	for (var i in this.files) {
		if (r.test(i)) {
			delete this.files[i];
		}
	};
};

Bundle.prototype.getBundle = function() {
	var ret = hereDoc(function(){/*!
;// Packed by nodebundle.js
;(function() {
	
	var _nodebundle_definitions = {};
	
	var _nodebundle_curdir = ["/virtual"];
	
	var _nodebundle_require = function(path) {
		
		var exec = function(str) {
			_nodebundle_curdir.push( str.substr(0, str.lastIndexOf("/")) );
			var module = {
				'exports': { }
			};
			_nodebundle_definitions[str](
				_nodebundle_require,
				module,
				module.exports,
				_nodebundle_curdir[_nodebundle_curdir.length-1],
				str
			);
			_nodebundle_curdir.pop();
			return module.exports;
		};
		
		var normalisePath = function(str) {
			str = str.split("//").join("/");               // Remove (most) //
			str = str.split("/./").join("/");              // Remove /./
			str = str.split(/\/[^\/]*\/..\//).join("/");   // Remove /x/../
			return str;
		};
		
		// Simulate the original require() algorithm. There is an overview of
		//  resolve order on http://nodejs.org/api/modules.html#modules_all_together
		
		// Attempt to load system module
		
		if (path[0] !== "/") {
			try {
				return require(path);
			} catch( e ) { }
		}
		
		// Attempt to load exact path
		
		if (path in _nodebundle_definitions) {
			return exec(path);		
		}
		
		// Attempt to load relative file
		
		var relpath = normalisePath(_nodebundle_curdir[_nodebundle_curdir.length-1] + "/" + path);
		if (relpath in _nodebundle_definitions) {
			return exec(relpath);
		}
		
		// TODO: Attempt to load node_modules/{identifier}/{package.json:main}
		
		// Attempt to load node_modules/{identifier}/index.js
		
		var indexpath = normalisePath(_nodebundle_curdir[_nodebundle_curdir.length-1] + "/node_modules/" + path + "/index.js");
		if (indexpath in _nodebundle_definitions) {
			return exec(indexpath);
		}
		
		// No way of loading the module
		
		throw new Error("Unable to load "+path);
		
	};
	
*/});
	
	// Truncate longest common prefix from files.
	var lcp = longestCommonPrefix(Object.keys(this.files));
	
	var jsident = function(sz) {
		return JSON.stringify( "/virtual/" + sz.substr(lcp.length) );
	};
	
	for (var i in this.files) {
		
		ret += (
			"\t_nodebundle_definitions[" + jsident(i) + "] = function(require, module, exports, __dirname, __filename){\n"+
			this.files[i]+
			"\n\t};\n"
		);
		
	}
	
	for (var i in this.execute) {
		ret += "\t_nodebundle_require(" + jsident(this.execute[i]) + ");\n";
	}
	
	ret += "\n})();\n";
	
	return ret;
};

Bundle.prototype.normalisePath = function(filepath) {
	return filepath.split("\\").join("/");
};

//
// Console entry
//

var usage = function() {
	console.log(
		"Usage: nodebundle [options]\n"+
		"\n"+
		"   -d {directory}       Add all files in a directory to bundle\n"+
		"   -e {file}            Execute file in bundle when bundle is executed\n"+
		"   -f {file}            Add a file to bundle\n"+
		"   -h,--help            Display this message\n"+
		"   -o {file}            Set output file. Use - for stdout\n"+
		"   -x {regexp}          Prune all added files matching regexp\n"
	);
};

var main = function(argv) {
	var b = new Bundle();
	var outpath = null;
	
	argv.push(null);
	
	for(var i = 0; i !== argv.length-1; ++i) {
		switch(argv[i]) {
			
			case '-d': {
				b.addDir( path.resolve(argv[++i]) );
			} break;
		
			case '-f': {
				b.addFile( path.resolve(argv[++i]) );
			} break;
		
			case '-e': {
				b.addExecute( path.resolve(argv[++i]) );
			} break;
		
			case '-o': {
				outpath = argv[++i];
			} break;
			
			case '-x': {
				b.pruneByRegexString( argv[++i] );
			} break;
		
			case '-h':
			case '--help':
			{
				usage();
				return;
			} break;
		}
	}
	
	if (outpath === null) {
		console.log("[nodebundle]: Output path not set\n");
		usage();
		
	} else if (outpath === "-") {
		console.log(b.getBundle());
		
	} else {
		fs.writeFileSync(outpath, b.getBundle());
		
	}
};

main(process.argv.slice(1));
